#!/usr/bin/env python3
# calculates the proton spectrum from the apparatus spectrum
import logging
from argparse import FileType, ArgumentDefaultsHelpFormatter, ArgumentParser

import numpy as np
from numpy import power, exp


log = logging.getLogger("mkproton")


# load calibration data from file
def read_calibration(calibration_file):
    elements = calibration_file.read().split()
    if len(elements) != 6:
        raise RuntimeError("calibration file does not contain 6 elements")
    cal = dict(time=int(elements[0]), first_used=int(elements[1]),
               ch1=int(elements[2]), e1=float(elements[3]),
               ch2=int(elements[4]), e2=float(elements[5]))
    cal['slope'] = (cal['e2'] - cal['e1']) / (cal['ch2'] - cal['ch1'])
    cal['intercept'] = cal['e1'] - cal['slope'] * cal['ch1']
    return cal


# generates the proton spectrum based on the apparatus spectrum and calibration
def proton_spectrum(calibration, apparatus_spectrum, step_size,
                    photon_yield_func=lambda e: e):
    min_band_ix, max_band_ix = get_band_limits(calibration,
                                               photon_yield_func,
                                               step_size)
    band_energy_bounds = [((band_ix - 1) * step_size, band_ix * step_size)
                          for band_ix in range(min_band_ix, max_band_ix + 1)]
    bands = [get_spectral_band(calibration, apparatus_spectrum,
                               photon_yield_func(l),
                               photon_yield_func(h))
             for (l, h) in band_energy_bounds]

    upper_limit = max_band_ix
    for i, band in enumerate(bands):  # discard bands after dropping to zero
        if np.isclose(band, 0.0, 1e-5):
            upper_limit = min_band_ix + (i - 1)
            break

    return bands, min_band_ix, upper_limit


def channel_to_energy(cal, channel):
    return cal['slope'] * channel + cal['intercept']


def energy_to_channel(cal, energy):
    return (energy - cal['intercept']) / cal['slope']


# predefined photon yield functions for stilbene scintillators
photon_yield_funcs = {
    'stil10': (lambda proton_energy:
               0.184318 * power(proton_energy, 1.39796)
               * exp(-.196851 * power(proton_energy, -1.5))),
    'stil20': (lambda proton_energy:
               0.16 * power(proton_energy, 1.48293)
               * exp(-0.141296 * power(proton_energy, -1.5) +
                     -0.00859061 * power(proton_energy, 0.9))),
    'stil45': (lambda proton_energy:
               0.151923 * power(proton_energy, 1.54268)
               * exp(-0.0829235 * power(proton_energy, -1.5) +
                     -0.0260154 * power(proton_energy, 0.9))),
    'photon': (lambda proton_energy: proton_energy)
}


# get valid proton spectrum indices
def get_band_limits(cal, photon_yield_func, step_size):
    first_band = None
    last_band = None

    min_photon_energy = channel_to_energy(cal, cal['first_used'])
    max_photon_energy = channel_to_energy(cal, 4096)

    for i in range(1, 4096):
        proton_energy = i * step_size
        photon_energy = photon_yield_func(proton_energy)

        if first_band is None and photon_energy > min_photon_energy:
            first_band = i + 1

        if photon_energy < max_photon_energy:
            last_band = i
        if photon_energy is not None and photon_energy > max_photon_energy:
            break

    assert 4095 > max_photon_energy > min_photon_energy
    return first_band, last_band


# integrate apparatus spectrum over the specified photon energy range
def get_spectral_band(cal, apparatus_spectrum,
                      min_photon_energy, max_photon_energy):
    from_channel = energy_to_channel(cal, min_photon_energy)
    to_channel = energy_to_channel(cal, max_photon_energy)

    from_idx = int(np.round(from_channel))
    to_idx = int(np.round(to_channel))

    assert to_idx > from_idx

    l = ((from_idx - from_channel + .5) * apparatus_spectrum[from_idx - 1]
         if from_idx - 1 < len(apparatus_spectrum) else 0)
    r = ((to_channel - to_idx + .5) * apparatus_spectrum[to_idx - 1]
         if to_idx - 1 < len(apparatus_spectrum) else 0)

    return l + np.sum(apparatus_spectrum[from_idx:to_idx - 1]) + r


def crystalsize_to_area(diameter):
    return np.pi * (diameter / 2000.) ** 2


def main():
    p = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                       description="converts apparatus to proton spectrum")
    p.add_argument('CALIBRATION',
                   type=FileType("r"))
    p.add_argument("APPARATUS_SPECTRUM",
                   type=FileType("r"),
                   help="file containing the apparatus spectrum")
    p.add_argument("STEP", type=float,
                   help="output proton spectrum step size")
    yield_args = p.add_mutually_exclusive_group(required=True)
    yield_args.add_argument('-y', '--yieldfunc',
                            choices=tuple(sorted(photon_yield_funcs.keys())),
                            help="use one of built-in photon yield functions")
    yield_args.add_argument('-t', '--yieldfile',
                            type=FileType('r'),
                            help=("load photon yield function as two column "
                                  + "text table, whitespace separated"))
    p.add_argument('-s', '--crystalsize',
                   type=int, default='10',
                   help="size of the scintillator used")
    p.add_argument("OUTPUT_FILE", nargs="?",
                   default="-",
                   type=FileType("w"))
    p.add_argument("-a", "--all", action='store_true',
                   help="ignore the upper limit and store all bands")
    args = p.parse_args()

    if args.yieldfunc:
        photon_yield_func = photon_yield_funcs[args.yieldfunc]
    else:
        try:
            import scipy.interpolate
        except ImportError:
            log.fatal("please install Scipy ($ pip install scipy) to use " +
                      "the photon yield function table loader")
            return 1
        table = np.loadtxt(args.yieldfile).T
        spline = scipy.interpolate.UnivariateSpline(x=table[0],
                                                    y=table[1],
                                                    k=1,  # linear
                                                    ext='raise')  # no magic
        photon_yield_func = spline

    calibration = read_calibration(args.CALIBRATION)
    apparatus_spectrum = np.loadtxt(args.APPARATUS_SPECTRUM)
    spectrum, min_band, max_band = proton_spectrum(calibration,
                                                   apparatus_spectrum,
                                                   args.STEP,
                                                   photon_yield_func)
    if args.all:
        max_band = min_band + len(spectrum) - 1
    else:
        spectrum = spectrum[:max_band - min_band + 1]

    print("#min", min_band, file=args.OUTPUT_FILE)
    print("#max", max_band, file=args.OUTPUT_FILE)
    print("#time", calibration['time'], file=args.OUTPUT_FILE)
    print("#crystal_area", crystalsize_to_area(args.crystalsize),
          file=args.OUTPUT_FILE)
    print("#step", args.STEP, file=args.OUTPUT_FILE)

    for i, band in enumerate(spectrum):
        print(band, file=args.OUTPUT_FILE)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    exit(main())


