#!/usr/bin/env python3
import argparse
import logging

import numpy as np

log = logging.getLogger("unfold")


def read_response_function(f):
    raw_elements = np.fromfile(f, dtype=np.double, sep='\n')
    matrix_size = int(np.sqrt(len(raw_elements)))
    if matrix_size ** 2 != len(raw_elements):
        raise ValueError("response matrix not square ({} elements)"
                         .format(len(raw_elements)))
    return raw_elements.reshape((matrix_size, matrix_size))


def unfold(proton_spectrum, response_matrix, min_band, maxiters, stop_diff=0):
    log.info("response matrix with {} bands loaded".format(response_matrix.shape[0]))
    response_matrix = np.asmatrix(response_matrix)
    top_limit = min(response_matrix.shape[1] - 1, min_band + len(proton_spectrum) - 1)
    response_cutout = response_matrix[min_band-1:top_limit, min_band-1:top_limit]
    response_norms = np.linalg.norm(response_cutout, 1, axis=0)[None, :]
    response_normed = response_cutout / response_norms
    proton_cutout = np.asarray(proton_spectrum[:(top_limit - min_band) + 1])
    spectrum_normed = proton_cutout / np.linalg.norm(proton_cutout, 1)

    result = np.ones_like(spectrum_normed) / len(spectrum_normed)

    def iterate():
        e = result * response_normed.T
        f_tmp = ((spectrum_normed / e) * response_normed)
        return np.asarray(f_tmp) * result

    def stop_value(aux, n):
        return np.linalg.norm(n - aux, 2) / np.linalg.norm(aux, 2)

    log.info("unfolding initialized, starting E-M")
    for i in range(maxiters):
        new_result = iterate()
        diff = stop_value(result, new_result)
        result = new_result[:]
        log.info("iteration {}: difference: {}".format(i + 1, diff))
        if diff < stop_diff:
            log.info("iteration difference dropped below limit")
            break
    return (result * sum(proton_cutout)) / response_norms


def normalize(neutron_spectrum, proton_spectrum, matrix_norms, step, time, crystal_area):
    normalized = divided / (step**2 * crystal_area * time)
    return normalized


def read_proton_spectrum(spectrum_file):
    meta, spectrum = {}, []
    tokens = spectrum_file.read().split()
    while tokens:
        if tokens[0].startswith("#"):
            if len(tokens[0]) < 2:
                raise RuntimeError("no control word after #")
            if len(tokens) < 2:
                raise RuntimeError("control word without value: " + tokens[0])
            meta[tokens[0][1:]] = tokens[1]
            tokens = tokens[2:]
        else:
            spectrum.append(float(tokens[0]))
            tokens = tokens[1:]

    if len(spectrum) < 2 or 'min' not in meta or 'max' not in meta:
        raise RuntimeError("invalid proton spectrum file")
    if int(meta['max']) - int(meta['min']) + 1 != len(spectrum):
        raise RuntimeError("invalid proton spectrum size")
    return spectrum, meta


def main():
    logging.basicConfig(level=logging.INFO)
    p = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    p.add_argument("response_function",
                   type=argparse.FileType("r"),
                   help="file containing the proton spectrum")
    p.add_argument("proton_spectrum",
                   type=argparse.FileType("r"),
                   help="file containing the response matrix")
    p.add_argument("iterations",
                   type=int, default=10, nargs='?',
                   help='maximum number of iterations')
    p.add_argument("limit",
                   type=float, default=0., nargs='?',
                   help='stop when the iteration difference drops below limit')
    args = p.parse_args()

    response_function = read_response_function(args.response_function)
    proton_spectrum, meta = read_proton_spectrum(args.proton_spectrum)

    l, h, time = int(meta['min']), int(meta['max']), int(meta['time'])
    crystal_area = float(meta.get('crystal_area', np.pi * 2.25e-2 ** 2))
    step = float(meta.get('step', 0.1))

    neutron_spectrum = unfold(proton_spectrum, response_function, l,
                              args.iterations, args.limit)

    flux = neutron_spectrum / (step**2 * crystal_area * time)

    def energies():
        return [(i - 0.5) * step for i in range(l, l + len(proton_spectrum))]

    for e, f in zip(energies(), flux.flatten().tolist()):
        print(e, f, sep='\t')


if __name__ == '__main__':
    exit(main())
